package Plugins::MSAD;
use strict;
use warnings;
use Config::General;
use MIME::Base64;
use Plugins::MSAD_core;
my (@f_attribs, $att, $val);

my $VERSION     = 2.2;
my $plugin_name = 'MSAD';
my $this_pkg    = (caller(0))[0] . '::' . $plugin_name;

my $conf = Config::General->new(
    -ConfigFile      => $config::plugins_dir . '/' . $plugin_name . '.conf',
    -InterPolateVars => 1,
    -AutoLaunder     => 1
);
my %cfg = $conf->getall;

foreach my $base_dn (keys %{ $cfg{'base_dn'} }) {
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: getting admin list: ' . join(", ", keys %{ $cfg{'base_dn'}->{$base_dn}->{'admin_users'} }))
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log(
        '-=- ' . $this_pkg . '[' . $base_dn . ']: getting exclusion list: ' . join(", ", keys %{ $cfg{'base_dn'}->{$base_dn}->{'exclude_users'} }))
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: users disabled=' . $cfg{'base_dn'}->{$base_dn}->{'AD_new_users_disabled'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: users password never expires=' . $cfg{'base_dn'}->{$base_dn}->{'AD_new_users_pne'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: users password expires now=' . $cfg{'base_dn'}->{$base_dn}->{'AD_new_users_pwd_exp'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: use custom suffix=' . $cfg{'base_dn'}->{$base_dn}->{'use_custom_suffix'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: custom suffix=' . $cfg{'base_dn'}->{$base_dn}->{'AD_custom_suffix'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: panic on errors=' . $cfg{'base_dn'}->{$base_dn}->{'panic_on_errors'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: show warnings as errors=' . $cfg{'base_dn'}->{$base_dn}->{'show_warnings_as_errors'})
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: modifyable attributes: ' . join(", ", keys %{ $cfg{'base_dn'}->{$base_dn}->{'m_attributes'} }))
      if $config::aosi_debug_level > 1;
    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: custom filter operator=' . $cfg{'base_dn'}->{$base_dn}->{'custom_filter_operator'})
      if $config::aosi_debug_level > 1;

    $cfg{'base_dn'}->{$base_dn}->{'filters'} = {};
    if ($cfg{'base_dn'}->{$base_dn}->{'custom_filter'}) {
        AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: custom filter=' . $cfg{'base_dn'}->{$base_dn}->{'custom_filter'})
          if $config::aosi_debug_level > 1;
        @f_attribs = split(/[,;]/, $cfg{'base_dn'}->{$base_dn}->{'custom_filter'});
        foreach my $f_attrib (@f_attribs) {
            ($att, $val) = split(/=/, $f_attrib);
            if (not exists $cfg{'base_dn'}->{$base_dn}->{'filters'}->{$att}) {
                $cfg{'base_dn'}->{$base_dn}->{'filters'}->{$att} = [];
            }
            push(@{ $cfg{'base_dn'}->{$base_dn}->{'filters'}->{$att} }, $val);
        }
        AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: filters=' . AOSICFG::dump_hash($cfg{'base_dn'}->{$base_dn}->{'filters'}))
          if $config::aosi_debug_level > 1;
    }

    if ($cfg{'base_dn'}->{$base_dn}->{'AD_profile_path'}) {
        AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: profile path=' . $cfg{'base_dn'}->{$base_dn}->{'AD_profile_path'})
          if $config::aosi_debug_level > 1;
    }

    AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: new users group='.$cfg{'base_dn'}->{$base_dn}->{'AD_new_dn'}) if $config::aosi_debug_level > 1;
    $cfg{'base_dn'}->{$base_dn}->{'ous'} = {};
    foreach my $att (keys %{ $cfg{'base_dn'}->{$base_dn}->{'users_ou'}}){
        AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']: users_ou for attribute '.$att) if $config::aosi_debug_level > 1;
        foreach my $val (keys %{ $cfg{'base_dn'}->{$base_dn}->{'users_ou'}->{$att}}){
            my $ou = $cfg{'base_dn'}->{$base_dn}->{'users_ou'}->{$att}->{$val};
            $cfg{'base_dn'}->{$base_dn}->{'ous'}->{$att}->{$val} = $ou;
            AOSICFG::write_ws_log('-=- ' . $this_pkg . '[' . $base_dn . ']:    mapping('.$att.'): '.$val.' to '.$ou) if $config::aosi_debug_level > 1;
        }
    }
}

sub beforeAddUser {
    my ($status_hash, $ldap, $admin_base_dn, $entry_in, $user, $pwd) = @_;
    my ($this_fun, $code, $msg, $out_panic, $msg2, $msg3, $filter);

    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;base_dn=$admin_base_dn;entry_in=$entry_in;") if $config::aosi_debug_level > 1;

    ($code, $msg, $out_panic, $msg2, $msg3) = Plugins::MSAD_core::beforeAddUser($status_hash, $ldap, $admin_base_dn, $entry_in, $user, $pwd, \%cfg);

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic, $msg2, $msg3;
}

##################################################################
#  Add an account
##################################################################
sub afterAddUser {
    my ($status_hash, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $code, $msg, $out_panic);

    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    ($code, $msg, $out_panic) = Plugins::MSAD_core::afterAddUser($status_hash, $entry_in, $out_xml, $user, $pwd, \%cfg, $plugin_name);

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
# Check if an account exists
##################################################################
sub beforeDeleteUser {
    my ($status_hash, $ldap, $admin_base_dn, $entry_in, $user, $pwd) = @_;
    my ($this_fun, $code, $msg, $out_panic, $msg2, $msg3);

    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;

    ($code, $msg, $out_panic, $msg2, $msg3) = Plugins::MSAD_core::beforeDeleteUser($status_hash, $ldap, $admin_base_dn, $entry_in, $user, $pwd, \%cfg);

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic, $msg2, $msg3;
}

##################################################################
# Delete an account
##################################################################
sub afterDeleteUser {
    my ($status_hash, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $code, $msg, $out_panic);
    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

    ($code, $msg, $out_panic) = Plugins::MSAD_core::afterDeleteUser($status_hash, $entry_in, $out_xml, $user, $pwd, \%cfg, $plugin_name);

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Check if adding an existig value or deleting a nonexisting value
##################################################################
# Same function for all actions
# 1 = beforeAddAttribute
# 2 = beforeModifyAttribute
# 3 = beforeDeleteAttribute
sub beforeChangeAttribute {
    my ($status_hash, $type, $ldap, $admin_base_dn, $entry_in, $user, $pwd) = @_;
    my ($this_fun, $code, $msg, $out_panic, $msg2, $msg3, $filter);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;ldap=$ldap;base=$admin_base_dn;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;

    ($code, $msg, $out_panic, $msg2, $msg3) = Plugins::MSAD_core::beforeChangeAttribute($status_hash, $type, $ldap, $admin_base_dn, $entry_in, $user, $pwd, \%cfg);

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic, $msg2, $msg3;
}


##################################################################
#  Change attributes (userPassword in AD is in fact unicodePwd)
##################################################################
# Same function for all actions
# 1 = afterAddAttribute
# 2 = afterModifyAttribute
# 3 = afterDeleteAttribute
sub afterChangeAttribute {
    my ($status_hash, $type, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $code, $msg, $out_panic);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    ($code, $msg, $out_panic) = Plugins::MSAD_core::afterChangeAttribute($status_hash, $type, $entry_in, $out_xml, $user, $pwd, \%cfg, $plugin_name);

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

1;
