package Plugins::MSAD_core;

use strict;
use warnings;
use MIME::Base64;
my ($att, $val);

my $VERSION     = 2.2;
my $core_name = 'MSAD_core';
my $this_pkg    = (caller(0))[0] . '::' . $core_name;

######################################################################
# Check if an uid is on the list of admins.
# Returns: a number of hits (usualy 0 or 1)
######################################################################
sub _item_is_in_list {
	my ($in, $list, $spec) = @_;
	my ($ok, $realm, $item, $this_fun);
	$this_fun = (caller(0))[3];

	$ok = 0;
	$item = $in;
	if((defined $spec) and ($spec == 1)){
		($item, $realm) = AOSICFG::user_split($in);
	}
	foreach my $member (@$list) {
		chomp($member);
		if ($member eq $item) {
			$ok++;
		}
	}
	AOSICFG::write_ws_log('|> ' . $this_fun . ': item ' . $in . ' is in list [' . join(", ", @$list) . ']: ' . $ok) if $config::aosi_debug_level > 2;
	return $ok;
}

######################################################################
# Wrapper for _user_is_ADadmin
######################################################################
sub user_is_ADadmin {
	my ($in, $base_dn, $cfg) = @_;
	my ($uid);

	$uid = $in;
	if($in =~/^cn=(.*)/){
		$uid = $1;
	}
	my @array = keys %{ $cfg->{'base_dn'}->{$base_dn}->{'admin_users'} };
	return _item_is_in_list($uid, \@array, 1);
}

######################################################################
# Check if the requiest is comming from direct access hosts
######################################################################
sub is_direct_access {
	my ($out, $this_fun);
	$this_fun = (caller(0))[3];

	$out = ($config::direct_access == 1) and (AOSICFG::ip_has_direct_access($AOSI::client_ip));
	AOSICFG::write_ws_log('|> ' . $this_fun . ': ' . $out) if $config::aosi_debug_level > 3;
	return $out;
}

######################################################################
# Check if an uid is in exclusion list
######################################################################
sub user_is_excluded {
	my ($in, $base_dn, $cfg) = @_;
	my @array = keys %{ $cfg->{'base_dn'}->{$base_dn}->{'exclude_users'} };
	return _item_is_in_list($in, \@array, 1);
}

######################################################################
# Check if an attribute is in modifiable list
######################################################################
sub can_modify_att {
	my ($in, $base_dn, $cfg) = @_;
	my @array = keys %{ $cfg->{'base_dn'}->{$base_dn}->{'m_attributes'} };
	return _item_is_in_list($in, \@array);
}

######################################################################
# Get the an uid (sAMAccountName) from ful DN
# Returns: uid
######################################################################
sub get_SAN {
	my ($dn) = @_;
	my (@parts, $cn, $uid);

	@parts = split(/,/, $dn);
	($cn, $uid) = split(/=/, $parts[0]);

	return $uid;
}

######################################################################
# Get the full AD DN for an uid (sAMAccountName)
# Retuns: DN (if found), error code and error message
######################################################################
sub get_ADDN {
	my ($ldap, $base, $uid) = @_;
	my ($code, $error, @ents, $req_dn, $this_fun, $aduid);
	$this_fun = (caller(0))[3];
	$req_dn = '';

	$aduid = $uid;
	if($uid =~/^cn=(.*)/){
		$aduid = $1;
	}
	$aduid =~ s/\s+//g;
	AOSICFG::write_ws_log("|> $this_fun(in): base=" . $base . ";uid=" . $uid . ";aduid=".$aduid.";") if $config::aosi_debug_level > 3;
	my $search = $ldap->search(
		base   => $base,
		filter => '(sAMAccountName=' . $aduid . ')',
		attrs  => ['distinguishedName'],
		scope  => 'sub'
	);
	$code  = $search->code;
	$error = $search->error;
	$error = sanitize($error);
	AOSICFG::write_ws_log("|> $this_fun(search result): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 3;

	if ($code == 0) {
		@ents = $search->entries;
		if ($#ents > -1) {
			$req_dn = $ents[0]->get('distinguishedName')->[0];
		}
		else {
			#if($uid eq 'ISVU'){
			#    $req_dn = $cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_dn'};
			#}
			#else {
			$code  = -10009;
			$error = 'User ' . $aduid . ' does not exist!';
			#}
		}
	}

	AOSICFG::write_ws_log("|> $this_fun(out): req_dn=" . $req_dn . ";code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 3;
	return $req_dn, $code, $error;
}

######################################################################
# Extract the uid part from DN
######################################################################
sub get_uid {
	my ($dn) = @_;

	my ($uid, $tmp) = split(/,/, $dn);
	($tmp, $uid) = split(/=/, $uid);
	return $uid;
}

######################################################################
# Extract base DN part from DN
######################################################################
sub get_basedn {
	my ($dn) = @_;
	my ($out, $base1, $base2);

	($out, $base1, $base2) = split(/,/, $dn);
	$out = $base1 . ',' . $base2;
	return $out;
}

######################################################################
# Extract base DN part from hrEduPersonUniqueID
######################################################################
sub get_user_basedn {
	my ($uqid) = @_;
	my ($uid, $base1, $base2, $out);

	($uid,   $base1) = split(/\@/, $uqid);
	($base1, $base2) = split(/\./, $base1);
	$out = 'dc=' . $base1 . ',dc=' . $base2;
	return $out;
}

######################################################################
# Convert a base DN to domain
# (eg. DC=intranet,DC=domain,DC=hr -> intranet.domain.hr)
######################################################################
sub basedn2domain {
	my ($basedn) = @_;
	my ($out, @temp);

	$out = $basedn;
	$out =~ s/DC=//ig;
	@temp = split(/,/, $out);
	$out = join(".", @temp);
	return $out;
}

##################################################################
# Check filter
##################################################################
sub ck_filter {
	my ($ldap, $base_dn, $entry_in, $user, $pwd, $new, $cfg) = @_;
	my ($this_fun, $code,  $msg,    $attrs, $a_name, $a_vals,   $f_vals,   $ok);
	my (@f_keys, $mesg,  $mesg2, $dn,     $uid,    @f_search, $s_filter, %options, $f_num);
	my ($entries_no, $entries2_no, $code2, $msg2);
	$this_fun = (caller(0))[3];
	$msg      = '';
	$code     = 0;
	$ok       = 0;
	$f_num    = scalar keys %{$cfg->{'base_dn'}->{$base_dn}->{'filters'}};

	AOSICFG::write_ws_log("|> $this_fun(in): base=$base_dn;user=$user;entry_in=$entry_in;new=$new;cfg=".join(", ", keys %{$cfg->{'base_dn'}->{$base_dn}})) if $config::aosi_debug_level > 1;
	@f_keys = keys %{$cfg->{'base_dn'}->{$base_dn}->{'filters'}};
	AOSICFG::write_ws_log("|> $this_fun: filters=".join(", ", @f_keys)) if $config::aosi_debug_level > 3;
	if ($#f_keys > -1) {
		if ($new == 1) {    # New user - check data in XML
			$attrs = $entry_in->{'attrs'};
			AOSICFG::write_ws_log("|> $this_fun(new): attrs=" . join(",", @$attrs) . ";") if $config::aosi_debug_level > 2;
			foreach my $attr (@$attrs) {
				$a_name = $attr->{'att'};
				$a_vals = $attr->{'val'};
				foreach my $a_val (@$a_vals) {
					foreach my $filter (keys %{$cfg->{'base_dn'}->{$base_dn}->{'filters'}}) {
						$f_vals = $cfg->{'base_dn'}->{$base_dn}->{'filters'}->{$filter};
						foreach my $f_val (@$f_vals) {
							AOSICFG::write_ws_log("|> $this_fun: $a_name ? $filter; $a_val ? $f_val;") if $config::aosi_debug_level > 3;
							if (($a_name eq $filter) and ($a_val eq $f_val)) {
								$ok++;
							}
						}
					}
				}
			}
			AOSICFG::write_ws_log("|> $this_fun: ok=$ok;f_num=$f_num;") if $config::aosi_debug_level > 3;
			if(lc($cfg->{'base_dn'}->{$base_dn}->{'custom_filter_operator'}) eq lc('and')){
				if($ok == $f_num){
					$code = 0;
				}
				else {
					#$code = error_code(-10069, $base_dn, $cfg);
					$code = -10069;
					$msg  = "User attributes do not meet the filter AND (" . $cfg->{'base_dn'}->{$base_dn}->{'custom_filter'} . ") requirements!";
				}
				AOSICFG::write_ws_log("|> $this_fun: filter operator is AND;msg=".$msg) if $config::aosi_debug_level > 3;
			}
			elsif(lc($cfg->{'base_dn'}->{$base_dn}->{'custom_filter_operator'}) eq lc('or')){
				if ($ok > 0) {
					$code = 0;
				}
				else {
					#$code = error_code(-10068, $base_dn, $cfg);
					$code = -10068;
					$msg  = "User attributes do not meet the filter OR (" . $cfg->{'base_dn'}->{$base_dn}->{'custom_filter'} . ") requirements!";
				}
				AOSICFG::write_ws_log("|> $this_fun: filter operator is OR;msg=".$msg) if $config::aosi_debug_level > 3;
			}
			else {
				#$code = error_code(-10067, $base_dn, $cfg);
				$code = -10067;
				$msg  = "Unknown filter operator: " . $cfg->{'base_dn'}->{$base_dn}->{'custom_filter_operator'} . "!";
			}
		}
		else {    # Old user - check data in LDAP
			$dn       = $entry_in->{'dn'};
			$uid      = get_uid($dn);
			@f_search = ();
			foreach my $filter (keys %{$cfg->{'base_dn'}->{$base_dn}->{'filters'}}) {
				$f_vals = $cfg->{'base_dn'}->{$base_dn}->{'filters'}->{$filter};
				foreach my $f_val (@$f_vals) {
					push(@f_search, $filter . '=' . $f_val);
				}
			}
			$s_filter = '(uid=' . $uid . ')';
			AOSICFG::write_ws_log("|> $this_fun: search filter1=$s_filter;") if $config::aosi_debug_level > 2;
			%options = (
				base      => $base_dn,
				filter    => $s_filter,
				attrs     => ['dn'],
				timelimit => 180
			);
			$mesg = $ldap->search(%options);
			$code = $mesg->code;
			$msg  = $mesg->error;
			$msg = sanitize($msg);
			$entries_no = -1;
			if(exists $mesg->{'entries'}){
				$entries_no = $#{ $mesg->{'entries'} };
			}
			AOSICFG::write_ws_log("|> $this_fun(looking for uid 1): uid=$uid;code=$code;msg=$msg;entries=$entries_no;") if $config::aosi_debug_level > 2;
			if (($code == 0) and ($entries_no > -1)) 
			{
				$s_filter = '(&(uid=' . $uid . ')(' . join(")(", @f_search) . '))';
				AOSICFG::write_ws_log("|> $this_fun: search filter2=$s_filter;") if $config::aosi_debug_level > 2;
				%options = (
					base      => $base_dn,
					filter    => $s_filter,
					attrs     => ['dn'],
					timelimit => 180
				);
				$mesg2 = $ldap->search(%options);
				$code2 = $mesg2->code;
				$msg2  = $mesg2->error;
				$entries2_no = -1;
				if(exists $mesg2->{'entries'}){
					$entries2_no = $#{ $mesg2->{'entries'} };
				}
				$msg2 = sanitize($msg2);
				AOSICFG::write_ws_log("|> $this_fun: code2=$code2;msg2=$msg2;") if $config::aosi_debug_level > 3;

				if ($code2 == 0) {
					AOSICFG::write_ws_log("|> $this_fun: entries2=" . $entries2_no ) if $config::aosi_debug_level > 3;
					if ($entries2_no == -1) {
						#$code = error_code(-10066, $base_dn, $cfg);
						$code = -10066;
						$msg  = "User attributes do not meet the filter (" . $cfg->{'base_dn'}->{$base_dn}->{'custom_filter'} . ") requirements!";
					}
				}
			}
		}
	}
	else {
		AOSICFG::write_ws_log("|> $this_fun: no filter") if $config::aosi_debug_level > 1;
	}
	AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg") if $config::aosi_debug_level > 1;
	return $code, $msg;
}

######################################################################
# Connect to AD, first as AOSIRO account to find out the realm
# AD DN for a user and then as the user
# Returns: AD LDAP object, error code and error message
######################################################################
sub AD_RO_connect {
	my ($user, $pwd, $check_admin, $base_dn, $cfg) = @_;
	my ($ad_ldap, $mesg_ro, $code_ro, $error_ro, $req_dn, $code, $msg, $error, $admin);
	my ($dec_pwd, $req, $uid, $realm, $this_fun, $out_panic);

	$out_panic = 0;
	$this_fun  = (caller(0))[3];
	$req_dn = '';

	AOSICFG::write_ws_log("|> $this_fun(in): user=$user; check_admin=$check_admin; base_dn=$base_dn") if $config::aosi_debug_level > 2;
	($uid, $realm) = split(/\@/, $user);
	if(($user eq '') and is_direct_access()){
		$uid = get_SAN($cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_dn'});
		$pwd = MIME::Base64::encode_base64($cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_pwd'});
		$user = $uid;
	}
	AOSICFG::write_ws_log("|> $this_fun: uid=$uid") if $config::aosi_debug_level > 2;
	AOSICFG::write_ws_log("|> $this_fun: AD at " . $cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} . ":" . $cfg->{'base_dn'}->{$base_dn}->{'AD_port'})
	if $config::aosi_debug_level > 2;
	$ad_ldap = Net::LDAP->new('ldaps://' . $cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'}, port => $cfg->{'base_dn'}->{$base_dn}->{'AD_port'}, timeout => 20);
	if ($ad_ldap) {
		$mesg_ro  = $ad_ldap->bind($cfg->{'base_dn'}->{$base_dn}->{'AD_aosiro_dn'}, password => $cfg->{'base_dn'}->{$base_dn}->{'AD_aosiro_pwd'});
		$code_ro  = $mesg_ro->code;
		$error_ro = $mesg_ro->error;
		#$error_ro =~ s/[\x00\n\r]//g;
		$error_ro = sanitize($error_ro);
		AOSICFG::write_ws_log("|> $this_fun: code=$code_ro;error=$error_ro;") if $config::aosi_debug_level > 2;
		if ($code_ro == 0) {
			($req_dn, $code, $error) = get_ADDN($ad_ldap, $cfg->{'base_dn'}->{$base_dn}->{'AD_base'}, $uid);
			AOSICFG::write_ws_log("|> $this_fun(AOSIRO search): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
			if ($code == 0) {
				if ($check_admin == 1) {
					$admin = user_is_ADadmin($user, $base_dn, $cfg);
					if ($admin > 0) {
						AOSICFG::write_ws_log("|> $this_fun(req DN): " . $req_dn);
						$dec_pwd = MIME::Base64::decode_base64($pwd);
						$req     = $ad_ldap->bind($req_dn, password => $dec_pwd);
						$code    = $req->code;
						$error   = $req->error;
						#$error =~ s/[\x00\n\r]//g;
						$error   = sanitize($error);
						AOSICFG::write_ws_log("|> $this_fun(req bind): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
						$msg = $error;
					}
					else {
						$msg  = 'User ' . $uid . ' has no admin privileges!';
						$code = -10003;
					}
				}
			}
			else {
				$msg       = 'AOSIRO user could not search for ' . $uid . ' (AD code: ' . $code . '; AD message: ' . $error . ').';
				$code      = -10002;
				$out_panic = $cfg->{'base_dn'}->{$base_dn}->{'panic_on_errors'};
			}
		}
		else {
			$msg       = 'AOSIRO user could not bind to AD (AD code: ' . $code_ro . '; AD message: ' . $error_ro . ').';
			$code      = -10001;
			$out_panic = $cfg->{'base_dn'}->{$base_dn}->{'panic_on_errors'};
		}
	}
	else {
		$code = -10000;
		$msg  = "Connection to MS AD server " . $cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} . ":" . $cfg->{'base_dn'}->{$base_dn}->{'AD_port'} . " failed: $@";
		$out_panic = $cfg->{'base_dn'}->{$base_dn}->{'panic_on_errors'};
	}
	return $ad_ldap, $code, $msg, $out_panic;
}

##################################################################
# Check if an account exists
##################################################################
sub beforeAddUser {
	my ($status_hash, $ldap, $admin_base_dn, $entry_in, $user, $pwd, $cfg) = @_;
	my ($this_fun, $ad_ldap, $mesg_ro,  $code_ro,   $error_ro, $base_dn, $msg2,  $msg3);
	my ($req_dn,   $code,    $msg,      $error,     $req,      $dec_pwd, $uid,     $realm, $search);
	my ($new_uid,  $new_dn,  $code_new, $error_new, $temp,     @entries, $admin,   $out_panic);
	my ($ent_attrs,$attr, 	 $att_name);
	$this_fun  = (caller(0))[3];
	$msg       = '';
	$code      = 0;
	$out_panic = 0;

	AOSICFG::write_ws_log("|> $this_fun(in): user=$user;base_dn=$admin_base_dn;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
	$base_dn = $admin_base_dn;
	$msg3    = $admin_base_dn;

	if ($cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} ne '') {
		($code, $msg) = ck_filter($ldap, $admin_base_dn, $entry_in, $user, $pwd, 1, $cfg);
		if ($code == 0) {
			$new_dn  = $entry_in->{'dn'};
			$new_uid = get_uid($new_dn);
			if (user_is_excluded($new_uid, $base_dn, $cfg)) {
				$code      = error_code(-10011, $base_dn, $cfg);
				$msg       = 'Uid ' . $new_uid . ' is on exclusion list!';
				$out_panic = 0;
			}
			else {
				($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1, $base_dn, $cfg);
				if ($code == 0) {
					$search = $ad_ldap->search(
						base   => $cfg->{'base_dn'}->{$base_dn}->{'AD_base'},
						filter => '(sAMAccountName=' . $new_uid . ')',
						attrs  => ['distinguishedName'],
						scope  => 'sub'
					);
					$code_new  = $search->code;
					$error_new = $search->error;
					#$error_new =~ s/[\x00\n\r]//g;
					$error = sanitize($error);
					AOSICFG::write_ws_log("|> $this_fun(looking for 2): uid=" . $new_uid . ";code=" . $code_new . ";error=" . $error_new)
					if $config::aosi_debug_level > 2;
					if ($code_new == 0) {
						@entries = $search->entries;
						AOSICFG::write_ws_log("|> $this_fun(user exists): uid=" . $new_uid . ";#entries=" . $#entries) if $config::aosi_debug_level > 2;
						if ($#entries > -1) {
							$code      = error_code(-10010, $base_dn, $cfg);
							$msg       = "User $new_uid allready exists!";
							$out_panic = $cfg->{'base_dn'}->{$base_dn}->{'panic_on_errors'};
						}
						else {
							$ent_attrs = $entry_in->{'attrs'};
							AOSICFG::write_ws_log("|> $this_fun: ent_attrs=[" . join(", ", @$ent_attrs) . "]") if $config::aosi_debug_level > 2;
							foreach my $attr (@$ent_attrs) {
								$att_name = $attr->{'att'};
								AOSICFG::write_ws_log("|> $this_fun: att=$att_name") if $config::aosi_debug_level > 2;
								if ($att_name eq 'userPassword') {
									$msg2 = encode_base64($attr->{'val'}->[0]);
								}
							}
							$msg = 'OK';
						}
					}
					else {
						$msg = '';
					}
				}
			}
		}
	}
	else {
		$code = 0;
		$msg  = "Skipping for base DN $base_dn!";
		AOSICFG::write_ws_log('|> ' . $this_fun . ': skipping for base_dn ' . $base_dn) if $config::aosi_debug_level > 1;
	}

	AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
	return $code, $msg, $out_panic, $msg2, $msg3;
}

##################################################################
#  Add an account
##################################################################
sub afterAddUser {
	my ($status_hash, $entry_in, $out_xml, $user, $pwd, $cfg, $plugin_name) = @_;
	my ($this_fun, $ad_ldap, $mesg_ro,  $code_ro,   $error_ro,  $uid,         $realm);
	my ($req_dn,   $code,    $msg,      $error,     $req,       $dec_pwd,     $search);
	my ($new_uid,  $new_dn,  $code_new, $error_new, $temp,      $admin);
	my ($dn,       $add_dn,  $attrs,    $ent_attrs, $out_panic, $before_code, $before_msg, $before_panic);
	my ($PWD,      $uqid,    $sn,       $givenName, $mesg_add,  $uac,         $mesg_mod);
	my ($msg1,     $error1,  $code1,    $npwd,      $msg2,      $error2,      $code2);
	my ($base_dn, $mesg_pwd, $orig_pwd, $mail,      $path,      $add_ou,    $att_val);
	my ($before_msg2, $before_msg3);

	$this_fun  = (caller(0))[3];
	$msg       = '';
	$code      = 0;
	$out_panic = 0;
	($uid, $realm) = split(/\@/, $user);

	AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

	($code, $msg) = Plugins::parse_ldap_xml($out_xml);
	AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

	$before_code  = $status_hash->{$plugin_name}->{'before'}->{'code'};
	$before_msg   = $status_hash->{$plugin_name}->{'before'}->{'msg'};
	$before_panic = $status_hash->{$plugin_name}->{'before'}->{'panic'};
	$before_msg2  = $status_hash->{$plugin_name}->{'before'}->{'msg2'};
	$before_msg3  = $status_hash->{$plugin_name}->{'before'}->{'msg3'};
	$base_dn      = $before_msg3;

	if ($cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} ne '') {
		if ($before_code == 0 and $before_panic == 0) {
			$orig_pwd = decode_base64($before_msg2);
			if ($code == 0) {    # This one is from parse_ldap_xml
				($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1, $base_dn, $cfg);
				if ($code == 0) {
					my $mesg = $ad_ldap->bind($cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_dn'}, password => $cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_pwd'});
					$code  = $mesg->code;
					$error = $mesg->error;
					#$error =~ s/[\x00\n\r]//g;
					$error = sanitize($error);
					AOSICFG::write_ws_log("|> $this_fun(AOSIWRT bind): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
					if ($code == 0) {
						$dn     = $entry_in->{'dn'};
						$uid    = get_uid($dn);
						$add_ou = $cfg->{'base_dn'}->{$base_dn}->{'AD_new_dn'};
						AOSICFG::write_ws_log("|> $this_fun: add_dn=" . $add_dn) if $config::aosi_debug_level > 1;
						$ent_attrs = $entry_in->{'attrs'};
						$attrs     = [];
						foreach my $attr (@$ent_attrs) {    # Fetch some attributes
							if (($attr->{'att'} eq 'userPassword')) {
								$PWD = $orig_pwd;
							}
							elsif (($attr->{'att'} eq 'hrEduPersonUniqueID')) {
								if ('ARRAY' eq ref($attr->{'val'})) {
									$uqid = join(",", @{ $attr->{'val'} });
								}
								else {
									$uqid = $attr->{'val'};
								}
								AOSICFG::write_ws_log("|> $this_fun: uqid=" . $uqid) if $config::aosi_debug_level > 1;
							}
							#elsif (($attr->{'att'} eq 'cn')) {
							#    $cn = join(" ", @{ $attr->{'val'} });
							#}
							elsif (($attr->{'att'} eq 'sn')) {
								$sn = join(" ", @{ $attr->{'val'} });
							}
							elsif (($attr->{'att'} eq 'mail')) {
								$mail = join(" ", @{ $attr->{'val'} });
							}
							elsif (($attr->{'att'} eq 'givenName')) {
								$givenName = join(" ", @{ $attr->{'val'} });
							}
							else {
								if ($attr->{'att'} !~ /^hrEduPerson/) {
									#push(@$attrs, $attr->{'att'} => $attr->{'val'});
								}
							}
							if(exists $cfg->{'base_dn'}->{$base_dn}->{'ous'}->{$attr->{'att'}}){
								$att_val = join(" ", @{ $attr->{'val'} });
								if(exists $cfg->{'base_dn'}->{$base_dn}->{'ous'}->{$attr->{'att'}}->{'default'}){
									$add_ou = $cfg->{'base_dn'}->{$base_dn}->{'ous'}->{$attr->{'att'}}->{'default'};
								}
								if(exists $cfg->{'base_dn'}->{$base_dn}->{'ous'}->{$attr->{'att'}}->{$att_val}){
									$add_ou = $cfg->{'base_dn'}->{$base_dn}->{'ous'}->{$attr->{'att'}}->{$att_val};
								}
							}
						}

						$add_dn = 'CN=' . $uid . ',' . $add_ou . ',' . $cfg->{'base_dn'}->{$base_dn}->{'AD_base'};
						AOSICFG::write_ws_log("|> $this_fun: new_dn=" . $add_dn) if $config::aosi_debug_level > 3;
						# Composing AD attributes
						push(@$attrs, 'cn'             => [$uid]);
						push(@$attrs, 'sAMAccountName' => [$uid]);
						push(@$attrs, 'uid'            => [$uid]);
						if ($cfg->{'base_dn'}->{$base_dn}->{'use_custom_suffix'} == 1) {
							push(@$attrs, 'userPrincipalName' => [ $uid . '@' . $cfg->{'base_dn'}->{$base_dn}->{'AD_custom_suffix'} ]);
						}
						else {
							push(@$attrs, 'userPrincipalName' => [ $uid . '@' . basedn2domain($cfg->{'base_dn'}->{$base_dn}->{'AD_base'}) ]);
						}
						push(@$attrs, 'displayName' => [ $givenName . ' ' . $sn ]);
						push(@$attrs, 'description' => [ 'hrEduPersonUniqueID: ' . $uqid ]);
						push(@$attrs, 'sn'          => [$sn]);
						push(@$attrs, 'givenName'   => [$givenName]);
						if ($mail ne '') {
							push(@$attrs, 'mail' => [$mail]);
						}
						$path = $cfg->{'base_dn'}->{$base_dn}->{'AD_profile_path'};
						if ($path ne '') {
							if (rindex($path, '\\') != (length($path) - 1)) {    # Fix the path
								$path .= '\\';
							}
							$path .= $uid;                                       # Append sAMAccountName to path
							push(@$attrs, 'profilePath' => [$path]);
						}
						push(@$attrs, 'objectClass' => [ 'top', 'person', 'user', 'organizationalPerson' ]);
						AOSICFG::write_ws_log("|> $this_fun: attrs=" . AOSICFG::dump_array($attrs)) if $config::aosi_debug_level > 3;
						$mesg_add = $ad_ldap->add($add_dn, attrs => $attrs);
						$code     = $mesg_add->code;
						$error    = $mesg_add->error;
						#$error =~ s/[\x00\n\r]//g;
						$error = sanitize($error);
						AOSICFG::write_ws_log("|> $this_fun(adding): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;

						# UserAccountControl flags
						# ACCOUNTDISABLE        0x0002      2
						# PASSWD_NOTREQD        0x0020      32
						# PASSWD_CANT_CHANGE    0x0040      64
						# NORMAL_ACCOUNT        0x0200      512
						# DONT_EXPIRE_PASSWORD  0x10000     65536
						# PASSWORD_EXPIRED      0x800000    8388608
						if ($code == 0) {
							$uac = 512;
							if ($cfg->{'base_dn'}->{$base_dn}->{'AD_new_users_disabled'} == 1) {
								$uac += 2;    # Set account is disabled
							}
							if ($cfg->{'base_dn'}->{$base_dn}->{'AD_new_users_pne'} == 1) {
								$uac += 65536;    # Set password never expires
							}
							if ($cfg->{'base_dn'}->{$base_dn}->{'AD_new_users_pwd_exp'} == 1) {
								$uac += 8388608;    # Set password expired
							}
							AOSICFG::write_ws_log("|> $this_fun(uac): $uac") if $config::aosi_debug_level > 2;
							$mesg_mod = $ad_ldap->modify($add_dn, replace => { 'userAccountControl' => $uac });
							$code1    = $mesg_mod->code;
							$error1   = $mesg_mod->error;
							#$error1 =~ s/[\x00\n\r]//g;
							$error1 = sanitize($error1);
							AOSICFG::write_ws_log("|> $this_fun(adding user): code=$code1;error=$error1") if $config::aosi_debug_level > 2;
							$msg1 = 'User ' . $uid . ' added but could not set properties (AD code: ' . $code1 . '; AD message: ' . $error1 . ').';

							# Set password
							map { $npwd .= "$_\000" } split(//, "\"$PWD\"");    # Encode to UTF-16
							$mesg_pwd = $ad_ldap->modify($add_dn, replace => { 'unicodePwd' => $npwd });
							$code2    = $mesg_pwd->code;
							$error2   = $mesg_pwd->error;
							#$error2 =~ s/[\x00\n\r]//g;
							$error2 = sanitize($error2);
							AOSICFG::write_ws_log("|> $this_fun(setting pwd): code=$code2;error=$error2") if $config::aosi_debug_level > 2;
							$msg2 = 'User ' . $uid . ' added but could not set password (AD code: ' . $code2 . '; AD message: ' . $error2 . ').';

							$code = $code1 + $code2;
							if ($code == 0) {
								$msg  = 'User ' . $uid . ' added (AD code: ' . $code . '; AD message: ' . $error . ').';
								$code = 0;
							}
							else {
								$msg  = $msg1 . "\n" . $msg2;
								$code = error_code(-10023, $base_dn, $cfg);
							}
						}
						else {
							$msg  = 'User ' . $uid . ' could not be added (AD code: ' . $code . '; AD message: ' . $error . ').';
							$code = -10022;
						}
					}
					else {
						$msg  = 'AOSIWRT user could not bind to AD (AD code: ' . $code . '; AD message: ' . $error . ').';
						$code = -10021;
					}
				}
			}
		}
		else {
			$code = error_code(-10020, $base_dn, $cfg);
			$msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
		}
	}
	else {
		$code = 0;
		$msg  = "Skipping for base DN $base_dn!";
		AOSICFG::write_ws_log('|> ' . $this_fun . ': skipping for base_dn ' . $base_dn) if $config::aosi_debug_level > 1;
	}
	AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
	return $code, $msg, $out_panic;
}

##################################################################
# Check if an account exists
##################################################################
sub beforeDeleteUser {
	my ($status_hash, $ldap, $admin_base_dn, $entry_in, $user, $pwd, $cfg) = @_;
	my ($this_fun, $ad_ldap, $mesg_ro, $code_ro, $error_ro, $msg, $search, $msg2, $msg3);
	my ($req_dn, $code, $error, $req, $dec_pwd, $uid, $realm, $base_dn);
	my ($del_uid, $del_dn, $code_del, $error_del, $temp, @entries, $out_panic);
	$this_fun  = (caller(0))[3];
	$msg       = '';
	$code      = 0;
	$out_panic = 0;
	$del_dn    = '';
	($uid, $realm) = split(/\@/, $user);

	AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
	$base_dn = $admin_base_dn;
	$msg3    = $admin_base_dn;

	if ($cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} ne '') {
		($code, $msg) = ck_filter($ldap, $admin_base_dn, $entry_in, $user, $pwd, 0, $cfg);
		if ($code == 0) {
			$del_dn  = $entry_in->{'dn'};
			$del_uid = get_uid($del_dn);
			AOSICFG::write_ws_log("|> $this_fun(uid): del_uid=" . $del_uid);
			if (user_is_excluded($del_uid, $base_dn, $cfg)) {
				$code = error_code(-10032, $base_dn, $cfg);
				$msg       = 'Uid ' . $del_uid . ' is in exclusion list!';
				$out_panic = 0;
			}
			else {
				($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1, $base_dn, $cfg);
				if ($code == 0) {
					$search = $ad_ldap->search(
						base   => $cfg->{'base_dn'}->{$base_dn}->{'AD_base'},
						filter => '(sAMAccountName=' . $del_uid . ')',
						attrs  => ['distinguishedName'],
						scope  => 'sub'
					);
					$code_del  = $search->code;
					$error_del = $search->error;
					#$error_del =~ s/[\x00\n\r]//g;
					$error_del = sanitize($error_del);
					AOSICFG::write_ws_log("|> $this_fun(looking for 3): uid=" . $del_uid . ";code=" . $code_del . ";error=" . $error_del)
					if $config::aosi_debug_level > 2;
					if ($code_del == 0) {
						@entries = $search->entries;
						if ($#entries == -1) {
							$code = error_code(-10031, $base_dn, $cfg);
							$msg  = "User $del_uid does not exists!";
						}
					}
					else {
						$msg  = 'Error searching for ' . $del_uid . ' (AD code: ' . $code_del . '; AD message: ' . $error_del . ').';
						$code = -10030;
					}
				}
			}
		}
	}
	else {
		$code = 0;
		$msg  = "Skipping for base DN $base_dn!";
		AOSICFG::write_ws_log('|> ' . $this_fun . ': skipping for base_dn ' . $base_dn) if $config::aosi_debug_level > 1;
	}

	AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
	return $code, $msg, $out_panic, $msg2, $msg3;
}

##################################################################
# Delete an account
##################################################################
sub afterDeleteUser {
	my ($status_hash, $entry_in, $out_xml, $user, $pwd, $cfg, $plugin_name) = @_;
	my ($this_fun, $ad_ldap, $mesg_ro, $code_ro, $error_ro);
	my ($req_dn, $code, $error, $req, $dec_pwd, $uid, $realm, $search);
	my ($new_uid,     $new_dn,     $code_new,    $error_new,   $temp,         $admin);
	my ($dn,          $del_dn,     $attrs,       $ent_attrs,   $mesg_del,     $out_panic);
	my ($before_code, $before_msg, $before_msg2, $before_msg3, $before_panic, $msg, $base_dn);
	$this_fun = (caller(0))[3];
	$msg      = '';
	$code     = 0;
	($uid, $realm) = split(/\@/, $user);

	AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

	($code, $msg) = Plugins::parse_ldap_xml($out_xml);
	AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

	$before_code  = $status_hash->{$plugin_name}->{'before'}->{'code'};
	$before_msg   = $status_hash->{$plugin_name}->{'before'}->{'msg'};
	$before_panic = $status_hash->{$plugin_name}->{'before'}->{'panic'};
	$before_msg3  = $status_hash->{$plugin_name}->{'before'}->{'msg3'};
	$base_dn      = $before_msg3;

	if ($cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} ne '') {
		if ($before_code == 0 and $before_panic == 0) {
			($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1, $base_dn, $cfg);
			if ($code == 0) {
				my $mesg = $ad_ldap->bind($cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_dn'}, password => $cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_pwd'});
				$code  = $mesg->code;
				$error = $mesg->error;
				#$error =~ s/[\x00\n\r]//g;
				$error = sanitize($error);
				AOSICFG::write_ws_log("|> $this_fun(AOSIWRT bind): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
				if ($code == 0) {
					$dn  = $entry_in->{'dn'};
					$uid = get_uid($dn);
					($del_dn, $code, $error) = get_ADDN($ad_ldap, $cfg->{'base_dn'}->{$base_dn}->{'AD_base'}, $uid);
					AOSICFG::write_ws_log("|> $this_fun: del_dn=" . $del_dn) if $config::aosi_debug_level > 1;
					if ($code == 0) {
						$mesg_del = $ad_ldap->delete($del_dn);
						$code     = $mesg_del->code;
						$error    = $mesg_del->error;
						#$error =~ s/[\x00\n\r]//g;
						$error = sanitize($error);
						AOSICFG::write_ws_log("|> $this_fun(deleting): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
						if ($code == 0) {
							$msg  = 'User ' . $uid . ' deleted (AD code: ' . $code . '; AD message: ' . $error . ').';
							$code = 0;
						}
						else {
							$msg  = 'User ' . $uid . ' could not be deleted (AD code: ' . $code . '; AD message: ' . $error . ').';
							$code = error_code(-10043, $base_dn, $cfg);
						}

					}
					else {
						$msg  = 'DN for user ' . $uid . ' could not be found (AD code: ' . $code . '; AD message: ' . $error . ').';
						$code = error_code(-10042, $base_dn, $cfg);
					}
				}
				else {
					$msg  = 'AOSIWRT user could not bind to AD (AD code: ' . $code . '; AD message: ' . $error . ').';
					$code = -10041;
				}
			}
		}
		else {
			$code = error_code(-10040, $base_dn, $cfg);
			$msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
		}
	}
	else {
		$code = 0;
		$msg  = "Skipping for base DN $base_dn!";
		AOSICFG::write_ws_log('|> ' . $this_fun . ': skipping for base_dn ' . $base_dn) if $config::aosi_debug_level > 1;
	}
	AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
	return $code, $msg, $out_panic;
}

##################################################################
#  Check if adding an existig value or deleting a nonexisting value
##################################################################
# Same function for all actions
# 1 = beforeAddAttribute
# 2 = beforeModifyAttribute
# 3 = beforeDeleteAttribute
sub beforeChangeAttribute {
	my ($status_hash, $type, $ldap, $admin_base_dn, $entry_in, $user, $pwd, $cfg) = @_;
	my ($this_fun, $msg, $code, $out_panic, $filter, $code_mod, $error_mod, $ad_ldap);
	my ($uid, $dn, $search, $att, $val, @entries, $user_uid, $realm, $msg2, $msg3, $temp);
	my ($base_dn, $attrs, $user_own_attr);
	$this_fun  = (caller(0))[3];
	$msg       = '';
	$code      = 0;
	$out_panic = 0;

	AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;ldap=$ldap;base=$admin_base_dn;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
	$base_dn = $admin_base_dn;
	$msg3    = $admin_base_dn;

	if ($cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} ne '') {
		($code, $msg) = ck_filter($ldap, $admin_base_dn, $entry_in, $user, $pwd, 0, $cfg);
		if ($code == 0) {
			$dn  = $entry_in->{'dn'};
			$uid = get_uid($dn);
			if (user_is_excluded($uid, $base_dn, $cfg)) {
				$code      = error_code(-10032, $base_dn, $cfg);
				$msg       = 'Uid ' . $uid . ' is in exclusion list!';
				$out_panic = 0;
			}
			else {
				$attrs = $entry_in->{'attrs'};
				foreach my $att_ref (@{$attrs}){
					if (ref($att_ref) eq 'HASH'){
						$att = $att_ref->{'att'};
						AOSICFG::write_ws_log("|> $this_fun: att=$att") if $config::aosi_debug_level > 2;


						#$att = $entry_in->{'attrs'}->[0]->{'att'};
						if(can_modify_att($att, $base_dn, $cfg)){
							($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 0, $base_dn, $cfg);
							if ($code == 0) {
								#$val = $entry_in->{'attrs'}->[0]->{'val'}->[0];
								$val = $att_ref->{'val'}->[0];
								AOSICFG::write_ws_log("|> $this_fun: att=$att;val=$val") if ($att ne 'userPassword') and ($config::aosi_debug_level > 2);
								($user_uid, $realm) = split(/\@/, $user);
								$user_own_attr = 0;
								if ( defined $user_uid ){
									AOSICFG::write_ws_log("|> $this_fun: user_uid=".$user_uid) if $config::aosi_debug_level > 2;
									$user_own_attr = $uid eq $user_uid;
								}
								# User changes his attributes OR admin changes attributes OR direct access is enabled
								if ($user_own_attr or (user_is_ADadmin($user, $base_dn, $cfg) > 0) or is_direct_access()) {
									AOSICFG::write_ws_log('|> ' . $this_fun . ': uid=' . $uid . ';user=' . $user . ';user_own_attr=' . $user_own_attr)
									if $config::aosi_debug_level > 2;
									$filter = '(&(sAMAccountName=' . $uid . ')';
									if ($att ne 'userPassword') {
										$filter .= '(' . $att . '=' . $val . ')';
									}
									else {
										$msg2 = encode_base64($val);
										AOSICFG::write_ws_log("|> $this_fun: att is pwd - storing to msg2") if ($att ne 'userPassword') and ($config::aosi_debug_level > 2);
									}
									$filter .= ')';
									$search = $ad_ldap->search(
										base   => $cfg->{'base_dn'}->{$base_dn}->{'AD_base'},
										filter => $filter,
										attrs  => [$att],
										scope  => 'sub'
									);
									$code_mod  = $search->code;
									$error_mod = $search->error;
									#$error_mod =~ s/[^\0-\x80]//g;
									$error_mod = sanitize($error_mod);
									AOSICFG::write_ws_log("|> $this_fun(looking for 4): filter=$filter;code=" . $code_mod . ";error=" . $error_mod)
									if $config::aosi_debug_level > 2;
									if ($code_mod == 0) {
										$code = 0;
										$msg  = 'Success';
										#$msg     = encode_base64($val);
										@entries = $search->entries;
										AOSICFG::write_ws_log("|> $this_fun(found): no_of_entries=" . $#entries) if $config::aosi_debug_level > 2;

										# Adding an existing value
										if (($type == 1) and ($#entries > -1)) {											
											AOSICFG::write_ws_log('|> ' . $this_fun . ': Attribute $att with value $val already exists!') if $config::aosi_debug_level > 1;
											$code      = error_code(-10051, $base_dn, $cfg);
											$msg       = "Attribute $att with value $val already exists!";
											$out_panic = $cfg->{'base_dn'}->{$base_dn}->{'panic_on_errors'};
										}

										# Deleting a nonexisting value
										if (($type == 3) and ($#entries == -1) and ($val ne '')) {
											AOSICFG::write_ws_log('|> ' . $this_fun . ': There is no attribute $att with value $val!') if $config::aosi_debug_level > 1;
											$code      = error_code(-10052, $base_dn, $cfg);
											$msg       = "There is no attribute $att with value $val!";
											$out_panic = $cfg->{'base_dn'}->{$base_dn}->{'panic_on_errors'};
										}
									}
								}
								else {
									AOSICFG::write_ws_log('|> ' . $this_fun . ': User can change only his attributes!') if $config::aosi_debug_level > 1;
								}
							}
						}
						else {
							AOSICFG::write_ws_log('|> ' . $this_fun . ': attribute '.$att.' is not on the modifiable attributes list!') if $config::aosi_debug_level > 1;
						}
					}
				}
			}
		}
	}
	else {
		$code = 0;
		$msg  = "Skipping for base DN $base_dn!";
		AOSICFG::write_ws_log('|> ' . $this_fun . ': skipping for base_dn ' . $base_dn) if $config::aosi_debug_level > 1;
	}
	AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
	return $code, $msg, $out_panic, $msg2, $msg3;
}

##################################################################
#  Change attributes (userPassword in AD is in fact unicodePwd)
##################################################################
# Same function for all actions
# 1 = afterAddAttribute
# 2 = afterModifyAttribute
# 3 = afterDeleteAttribute
sub afterChangeAttribute {
	my ($status_hash, $type, $entry_in, $out_xml, $user, $pwd, $cfg, $plugin_name) = @_;
	my ($this_fun,    $out_panic,  $msg,          $mesg_mod, $mod_dn,   $dec_pwd,   $code);
	my ($uid,         $dn,         $att,          $val,      $code_mod, $error_mod, @entries, $user_uid, $realm);
	my ($before_code, $before_msg, $before_panic, $ad_ldap,  $nval,     $error,     $base_dn);
	my ($before_msg2, $before_msg3,$attrs, $user_own_attr);
	$this_fun  = (caller(0))[3];
	$msg       = '';
	$code      = 0;
	$out_panic = 0;

	AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
	($code, $msg) = Plugins::parse_ldap_xml($out_xml);
	AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

	$before_code  = $status_hash->{$plugin_name}->{'before'}->{'code'};
	$before_msg   = $status_hash->{$plugin_name}->{'before'}->{'msg'};
	$before_panic = $status_hash->{$plugin_name}->{'before'}->{'panic'};
	$before_msg2  = $status_hash->{$plugin_name}->{'before'}->{'msg2'};
	$before_msg3  = $status_hash->{$plugin_name}->{'before'}->{'msg3'};
	$base_dn      = $before_msg3;

	if ($cfg->{'base_dn'}->{$base_dn}->{'AD_hostname'} ne '') {
		if ($before_code == 0 and $before_panic == 0) {
			($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 0, $base_dn, $cfg);
			if ($code == 0) {
				$dn  = $entry_in->{'dn'};
				$attrs = $entry_in->{'attrs'};
				foreach my $att_ref (@{$attrs}){
					if (ref($att_ref) eq 'HASH'){
						$att = $att_ref->{'att'};
						AOSICFG::write_ws_log("|> $this_fun: att=$att") if $config::aosi_debug_level > 2;
						if($att ne 'userPassword'){
							$val = $att_ref->{'val'}->[0];
							AOSICFG::write_ws_log("|> $this_fun: val=".$val) if $config::aosi_debug_level > 2;
						}
						$uid = get_uid($dn);
						($user_uid, $realm) = split(/\@/, $user);
						$user_own_attr = 0;
						if ( defined $user_uid ){
							AOSICFG::write_ws_log("|> $this_fun: user_uid=".$user_uid) if $config::aosi_debug_level > 2;
							$user_own_attr = $uid eq $user_uid;
						}
						# User changes his attributes OR admin changes attributes OR direct access is enabled
						if ($user_own_attr or (user_is_ADadmin($user, $base_dn, $cfg) > 0) or is_direct_access()) {
							AOSICFG::write_ws_log('|> ' . $this_fun . ': uid=' . $uid . ';user=' . $user . ';user_own_attr=' . $user_own_attr) if $config::aosi_debug_level > 2;
							if ($att eq 'userPassword') {
								$att = 'unicodePwd';
								$val = decode_base64($before_msg2);
								AOSICFG::write_ws_log("|> $this_fun: changing password") if $config::aosi_debug_level > 2;
								map { $nval .= "$_\000" } split(//, "\"$val\"");
								$val = $nval;
							}
							AOSICFG::write_ws_log("|> $this_fun: bind_as_admin") if $config::aosi_debug_level > 2;
							$mesg_mod = $ad_ldap->bind($cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_dn'}, password => $cfg->{'base_dn'}->{$base_dn}->{'AD_aosiwrt_pwd'});
							$code     = error_code($mesg_mod->code, $base_dn, $cfg);
							$msg      = $mesg_mod->error;
							#$msg =~ s/[\x00\n\r]//g;
							$msg = sanitize($msg);
							AOSICFG::write_ws_log("|> $this_fun(bind): code=" . $code . ";error=" . $msg) if $config::aosi_debug_level > 2;
							if ($code == 0) {
								($mod_dn, $code, $msg) = get_ADDN($ad_ldap, $cfg->{'base_dn'}->{$base_dn}->{'AD_base'}, $uid);
								if ($code == 0) {
									AOSICFG::write_ws_log("|> $this_fun(replacing): $att=$val for dn=$mod_dn") if $att ne 'unicodePwd';
									if ($type == 1 or $type == 2) {
										$mesg_mod = $ad_ldap->modify($mod_dn, replace => { $att => $val });
									}
									elsif ($type == 3) {
										$mesg_mod = $ad_ldap->modify($mod_dn, delete => { $att => $val });
									}
									$code = error_code($mesg_mod->code, $base_dn, $cfg);
									$msg  = $mesg_mod->error;
								}
								$code = error_code($code, $base_dn, $cfg);
								$msg  = sanitize($msg);
							}

						}
						else {
							$code = error_code(-10061, $base_dn, $cfg);
							$msg  = 'User can change only his attributes!';
						}
					}
				}
			} 
			else {
				$code = error_code(-10062, $base_dn, $cfg);
				$msg  = "Read-only user cannot bind to AD: code=" . $code . "; message=" . $msg;
			}
		}
		else {
			$code = error_code(-10060, $base_dn, $cfg);
			$msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
		}
	}
	else {
		$code = 0;
		$msg  = "Skipping for base DN $base_dn!";
		AOSICFG::write_ws_log('|> ' . $this_fun . ': skipping for base_dn ' . $base_dn) if $config::aosi_debug_level > 1;
	}
	AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
	return $code, $msg, $out_panic;
}

sub sanitize {
	my ($in, $mode) = @_;
	my($out);

	if(!defined $mode){
		$mode = 1;
	}
	$out = $in;
	$out =~ s/[\0]//g;    # remove EOS from AD
	$out =~ s/[\x00]//g;
	if($mode == 2){       # ascii only
		$out =~ s/[^[:ascii:]]//g;
	}

	return $out;
}

sub error_code {
	my($real_code, $base_dn, $cfg) = @_;
	my($out, $this_fun);
	$this_fun  = (caller(0))[3];

	if($cfg->{'base_dn'}->{$base_dn}->{'show_warnings_as_errors'} == 1){
		$out = $real_code;
	}
	else {
		$out = 0;
	}
	AOSICFG::write_ws_log('|> ' . $this_fun . ': show_warnings_as_errors=' . $cfg->{'base_dn'}->{$base_dn}->{'show_warnings_as_errors'} . ';in_code=' . $real_code . ';out_code=' . $out) if $config::aosi_debug_level > 3;
	return $out;
}

1;
